/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @format
 * @emails oncall+relay
 * @flow strict-local
 */

'use strict';

import type {readUpdatableQueryEXPERIMENTALTestRegularQuery} from './__generated__/readUpdatableQueryEXPERIMENTALTestRegularQuery.graphql';
import type {OpaqueScalarType} from './OpaqueScalarType';

const RelayNetwork = require('../../network/RelayNetwork');
const {graphql} = require('../../query/GraphQLTag');
const RelayModernEnvironment = require('../../store/RelayModernEnvironment');
const {
  createOperationDescriptor,
} = require('../../store/RelayModernOperationDescriptor');
const RelayModernStore = require('../../store/RelayModernStore');
const RelayReader = require('../../store/RelayReader');
const RelayRecordSource = require('../../store/RelayRecordSource');
const commitLocalUpdate = require('../commitLocalUpdate');
const {
  validate: validateNode,
} = require('./__generated__/readUpdatableQueryEXPERIMENTALTest_node.graphql');
const {
  validate: validateUser,
} = require('./__generated__/readUpdatableQueryEXPERIMENTALTest_user.graphql');
const {createOpaqueScalarTypeValue} = require('./OpaqueScalarType');
const {
  disallowConsoleErrors,
  disallowWarnings,
} = require('relay-test-utils-internal');

disallowWarnings();
disallowConsoleErrors();

const updatableQuery = graphql`
  query readUpdatableQueryEXPERIMENTALTestUpdatableQuery @updatable {
    me {
      __typename
      __id
      id
      name
      author {
        client_best_friend {
          ...readUpdatableQueryEXPERIMENTALTest_user
          name
        }
      }
      author2: author {
        client_nickname
      }
      ...readUpdatableQueryEXPERIMENTALTest_user
    }
    node(id: "4") {
      ... on User {
        __typename
        name
      }
    }
    node2: node(id: "5") {
      ... on User {
        __typename
        name
        parents {
          ...readUpdatableQueryEXPERIMENTALTest_user
          name
          parents {
            name
          }
        }
      }
    }
    node3: node(id: "6") {
      id
      ...readUpdatableQueryEXPERIMENTALTest_node
    }
  }
`;

const updatableQuery2 = graphql`
  query readUpdatableQueryEXPERIMENTALTest2UpdatableQuery($id: ID!) @updatable {
    node(id: $id) {
      __typename
    }
  }
`;

const regularQuery = graphql`
  query readUpdatableQueryEXPERIMENTALTestRegularQuery {
    me {
      ...readUpdatableQueryEXPERIMENTALTest_node
      id
      name
      author {
        client_best_friend {
          name
        }
        client_nickname
      }
    }
    node(id: "4") {
      ...readUpdatableQueryEXPERIMENTALTest_user
      ... on User {
        name
      }
    }
    node2: node(id: "5") {
      ... on User {
        name
        parents {
          id
          name
          parents {
            id
          }
        }
      }
    }
  }
`;

graphql`
  fragment readUpdatableQueryEXPERIMENTALTest_user on User @assignable {
    __typename
  }
`;

graphql`
  fragment readUpdatableQueryEXPERIMENTALTest_node on Node @assignable {
    __typename
  }
`;

describe('readUpdatableQuery', () => {
  let environment;
  let operation;
  let rootRequest;

  beforeEach(() => {
    rootRequest = regularQuery;
    operation = createOperationDescriptor(rootRequest, {});
    const source = RelayRecordSource.create();
    const store = new RelayModernStore(source);

    const fetch = jest.fn();
    environment = new RelayModernEnvironment({
      network: RelayNetwork.create(fetch),
      store,
    });
  });

  it('can be used to read scalar values', () => {
    environment.commitPayload(operation, {
      me: {
        id: '4',
        __typename: 'User',
        name: 'Zuck',
        author: null,
      },
      node: null,
      node2: null,
    });

    commitLocalUpdate(environment, store => {
      const me = store.getRoot()?.getLinkedRecord('me');
      expect(me?.getValue('id')).toEqual('4');
      expect(me?.getValue('name')).toEqual('Zuck');

      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      expect(updatableData.me?.id).toEqual('4');
      expect(updatableData.me?.name).toEqual('Zuck');
    });
  });

  it('can be used to update scalar values', () => {
    environment.commitPayload(operation, {
      me: {
        id: '4',
        __typename: 'User',
        name: 'Zuck',
        author: null,
      },
      node: null,
      node2: null,
    });

    commitLocalUpdate(environment, store => {
      const me = store.getRoot()?.getLinkedRecord('me');
      expect(me?.getValue('id')).toEqual('4');
      expect(me?.getValue('name')).toEqual('Zuck');

      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      if (updatableData.me != null) {
        updatableData.me.name = 'MetaZuck';
      }
      expect(updatableData.me?.name).toEqual('MetaZuck');
      expect(me?.getValue('name')).toEqual('MetaZuck');
    });

    const source = environment.getStore().getSource();
    const selector = operation.fragment;
    const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
      .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
    expect(readOnlyData?.me?.name).toEqual('MetaZuck');
  });

  it('cannot be used to update clientids, ids or typenames', () => {
    environment.commitPayload(operation, {
      me: {
        id: '4',
        __typename: 'User',
        name: 'Zuck',
        author: null,
      },
      node: null,
      node2: null,
    });

    commitLocalUpdate(environment, store => {
      const me = store.getRoot()?.getLinkedRecord('me');
      expect(me?.getValue('id')).toEqual('4');
      expect(me?.getValue('name')).toEqual('Zuck');

      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      expect(() => {
        if (updatableData.me != null) {
          // $FlowFixMe[cannot-write] That's the point!
          updatableData.me.id = '5';
        }
      }).toThrowError();
      expect(() => {
        if (updatableData.me != null) {
          // $FlowFixMe[cannot-write] That's the point!
          updatableData.me.__typename = 'Protoss';
        }
      }).toThrowError();
      expect(() => {
        if (updatableData.me != null) {
          // $FlowFixMe[cannot-write] That's the point!
          updatableData.me.__id = '5';
        }
      }).toThrowError();
    });
  });

  it('does not create setters for fields within non-matching inline fragments', () => {
    environment.commitPayload(operation, {
      me: null,
      node: {
        __typename: 'User',
        id: '4',
        name: 'Zuck',
      },
      node2: {
        __typename: 'Page',
        id: '5',
      },
    });

    commitLocalUpdate(environment, store => {
      const node = store.getRoot()?.getLinkedRecord('node', {id: '4'});
      expect(node?.getValue('id')).toEqual('4');
      expect(node?.getValue('name')).toEqual('Zuck');

      const node2 = store.getRoot()?.getLinkedRecord('node', {id: '5'});
      expect(node2?.getValue('id')).toEqual('5');
      expect(node2?.getValue('name')).toEqual(undefined);

      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      if (updatableData.node != null) {
        const propertyDescriptor = Object.getOwnPropertyDescriptor(
          updatableData.node,
          'name',
        );
        expect(propertyDescriptor?.set).not.toEqual(undefined);
      }
      if (updatableData.node != null) {
        if (updatableData.node.__typename === 'User') {
          updatableData.node.name = 'MetaZuck';
          expect(updatableData.node.name).toEqual('MetaZuck');
        } else {
          throw new Error('Expected updatableData.node to have type User');
        }
      }
      expect(node?.getValue('name')).toEqual('MetaZuck');

      // Because node2 is not a user, selections within an inline fragment on User do not result
      // in a setter being created
      if (updatableData.node2 != null) {
        const propertyDescriptor2 = Object.getOwnPropertyDescriptor(
          updatableData.node2,
          'name',
        );
        expect(propertyDescriptor2).toEqual(undefined);

        // In __DEV__, this throws. In prod, this updates a value on an object, but that
        // doesn't update any value in the store, as evidenced by the undefined property
        // descriptor.
        expect(() => {
          if (updatableData.node2 != null) {
            // $FlowFixMe[prop-missing] that's the point!
            updatableData.node2.name = 'MetaZuck';
          }
        }).toThrowError();
      } else {
        throw new Error('Expected node2 to be found');
      }
    });
  });

  describe('assignable fragments with concrete type conditions', () => {
    it('lets you assign values to linked fields', () => {
      environment.commitPayload(operation, {
        me: {
          __typename: 'User',
          id: '42',
          name: 'NotZuck',
          author: null,
        },
        node: {
          __typename: 'User',
          id: '4',
          name: 'Zuck',
        },
        node2: null,
      });

      commitLocalUpdate(environment, store => {
        const me = store.getRoot()?.getLinkedRecord('me');
        expect(me?.getValue('id')).toEqual('42');
        expect(me?.getValue('name')).toEqual('NotZuck');

        const node = store.getRoot()?.getLinkedRecord('node', {id: '4'});
        expect(node?.getValue('id')).toEqual('4');
        expect(node?.getValue('name')).toEqual('Zuck');

        const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
          updatableQuery,
          {},
        ).updatableData;

        const source = environment.getStore().getSource();
        const selector = operation.fragment;
        const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
          .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);

        expect(readOnlyData.me?.id).toBe('42');

        if (readOnlyData.node != null) {
          if (readOnlyData.node.__typename === 'User') {
            expect(readOnlyData.node.__id).toBe('4');
          }
        }

        if (updatableData.me != null) {
          if (readOnlyData.node != null) {
            const validUser = validateUser(readOnlyData.node);
            if (validUser !== false) {
              updatableData.me = validUser;
              // Flow inference must be cleared here
              (() => {})();
              expect(updatableData.me?.id).toBe('4');
            } else {
              throw new Error('Expected assignment to be valid');
            }
          } else {
            throw new Error('Expected readOnlyData.node to exist');
          }
        } else {
          throw new Error('Expected updatableData.me to exist');
        }
      });

      const source = environment.getStore().getSource();
      const selector = operation.fragment;
      const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
        .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
      expect(readOnlyData.me?.id).toBe('4');
    });

    it('lets you assign to plural linked fields', () => {
      environment.commitPayload(operation, {
        me: null,
        node: {
          __typename: 'User',
          id: '4',
          name: 'Gaius Julius Caesar',
        },
        node2: {
          __typename: 'User',
          id: '5',
          name: 'Gaius Julius Caesar Octavianus',
          parents: [
            {
              __typename: 'User',
              id: '467',
              name: 'Gaius Octavius',
              parents: [],
            },
          ],
        },
      });

      commitLocalUpdate(environment, store => {
        const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
          updatableQuery,
          {},
        ).updatableData;

        const source = environment.getStore().getSource();
        const selector = operation.fragment;
        const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
          .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);

        const validUser = (() => {
          if (readOnlyData.node != null) {
            if (readOnlyData.node.__typename === 'User') {
              const forAssignment = validateUser(readOnlyData.node);
              if (forAssignment !== false) {
                return forAssignment;
              }
            }
          }
          throw new Error(
            'Expected readOnlyData.node to exist, __typename User and be valid for assignment',
          );
        })();

        const node2 = (() => {
          if (updatableData.node2 != null) {
            if (updatableData.node2.__typename === 'User') {
              return updatableData.node2;
            }
          }
          throw new Error('Expected node2 to exist and have __typename User');
        })();

        node2.parents = [validUser];

        // We need to clear Flow's inference about what type node2.parents has
        (() => {})();

        if (node2.parents[0] != null) {
          expect(node2.parents[0]?.name).toBe('Gaius Julius Caesar');
        } else {
          throw new Error('Expected parent to exist');
        }
      });

      const source = environment.getStore().getSource();
      const selector = operation.fragment;
      const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
        .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
      if (readOnlyData.node2?.parents != null) {
        expect(readOnlyData.node2?.parents[0]?.name).toBe(
          'Gaius Julius Caesar',
        );
      } else {
        throw new Error('Expected parents to exist');
      }
    });

    it('lets you update client extension linked fields', () => {
      environment.commitPayload(operation, {
        me: {
          __typename: 'User',
          id: '4',
          name: 'Mark',
          author: {
            id: '5',
            client_best_friend: {
              id: '6',
              name: 'Sheryl',
            },
            client_nickname: 'Zucc',
          },
        },
        node: {
          id: '4',
          __typename: 'User',
          name: 'Mark',
        },
        node2: null,
      });

      commitLocalUpdate(environment, store => {
        const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
          updatableQuery,
          {},
        ).updatableData;

        const source = environment.getStore().getSource();
        const selector = operation.fragment;
        const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
          .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);

        const author = (() => {
          if (updatableData.me?.author != null) {
            return updatableData.me.author;
          }
          throw new Error('Expected author to exist');
        })();

        const validUser = (() => {
          if (readOnlyData.node != null) {
            const forAssignment = validateUser(readOnlyData.node);
            if (forAssignment !== false) {
              return forAssignment;
            }
          }
          throw new Error(
            'Expected readOnlyData.node to exist and be valid for assignment',
          );
        })();

        author.client_best_friend = validUser;
        expect(updatableData.me?.author?.client_best_friend?.name).toBe('Mark');
      });
      const source = environment.getStore().getSource();
      const selector = operation.fragment;
      const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
        .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
      expect(readOnlyData.me?.author?.client_best_friend?.name).toBe('Mark');
    });

    describe('validate', () => {
      it('will return false if the source has the wrong __typename', () => {
        environment.commitPayload(operation, {
          me: null,
          node: {
            __typename: 'Alien',
            id: '4',
          },
          node2: null,
        });

        const source = environment.getStore().getSource();
        const selector = operation.fragment;
        const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
          .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
        if (readOnlyData.node != null) {
          expect(validateUser(readOnlyData.node)).toBe(false);
        } else {
          throw new Error('Expected readOnlyData.node to exist');
        }
      });

      it('will return the parameter if the source has a matching __typename', () => {
        environment.commitPayload(operation, {
          me: null,
          node: {
            __typename: 'User',
            id: '4',
            name: 'Mark',
          },
          node2: null,
        });

        const source = environment.getStore().getSource();
        const selector = operation.fragment;
        const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
          .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
        if (readOnlyData.node != null) {
          expect(validateUser(readOnlyData.node)).toEqual(readOnlyData.node);
        } else {
          throw new Error('Expected readOnlyData.node to exist');
        }
      });
    });
  });

  describe('assignable fragments with abstract type conditions', () => {
    it('lets you assign values to linked fields', () => {
      environment.commitPayload(operation, {
        me: {
          __typename: 'User',
          id: '4',
          author: null,
          name: null,
        },
        node: null,
        node2: null,
      });

      commitLocalUpdate(environment, store => {
        const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
          updatableQuery,
          {},
        ).updatableData;

        const source = environment.getStore().getSource();
        const selector = operation.fragment;
        const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
          .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);

        const validNode = (() => {
          if (readOnlyData.me != null) {
            const forAssignment = validateNode(readOnlyData.me);
            if (forAssignment !== false) {
              return forAssignment;
            }
          }
          throw new Error('Expected me to be present and valid for assignment');
        })();

        updatableData.node3 = validNode;
        (() => {})();
        expect(updatableData.node3?.id).toBe('4');
      });
    });

    describe('validate', () => {
      it('will return the parameter if the source has a matching __isFragmentName field', () => {
        environment.commitPayload(operation, {
          me: {
            __typename: 'User',
            id: '4',
            name: 'Mark',
            author: null,
            __isreadUpdatableQueryEXPERIMENTALTest_node: null,
          },
          node: null,
          node2: null,
        });

        const source = environment.getStore().getSource();
        const selector = operation.fragment;
        const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
          .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
        expect(
          readOnlyData.me?.__isreadUpdatableQueryEXPERIMENTALTest_node,
        ).toBe('User');
        if (readOnlyData.me != null) {
          expect(validateNode(readOnlyData.me)).toBe(readOnlyData.me);
        } else {
          throw new Error('Expected readOnlyData.me to exist');
        }
      });
    });
  });

  it('lets you delete linked fields by assigning null', () => {
    environment.commitPayload(operation, {
      me: {
        __typename: 'User',
        id: '42',
        name: 'NotZuck',
        author: null,
      },
      node: {
        __typename: 'User',
        id: '4',
        name: 'Zuck',
      },
      node2: null,
    });

    commitLocalUpdate(environment, store => {
      const me = store.getRoot()?.getLinkedRecord('me');
      expect(me?.getValue('id')).toEqual('42');
      expect(me?.getValue('name')).toEqual('NotZuck');

      const node = store.getRoot()?.getLinkedRecord('node', {id: '4'});
      expect(node?.getValue('id')).toEqual('4');
      expect(node?.getValue('name')).toEqual('Zuck');

      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      if (updatableData.me != null) {
        updatableData.me = null;
      }

      expect(updatableData.me).toBe(null);
      expect(store.getRoot()?.getLinkedRecord('me')).toBe(null);
    });

    const source = environment.getStore().getSource();
    const selector = operation.fragment;
    const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
      .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
    expect(readOnlyData.me).toBe(null);
  });

  it('allows you to assign an empty array to clear plural linked fields', () => {
    environment.commitPayload(operation, {
      me: null,
      node: null,
      node2: {
        __typename: 'User',
        id: '5',
        name: 'Oedipus',
        parents: [
          {
            __typename: 'User',
            id: '467',
            name: 'Laius',
            parents: [],
          },
        ],
      },
    });

    commitLocalUpdate(environment, store => {
      const parents = store
        .getRoot()
        .getLinkedRecord('node', {id: '5'})
        ?.getLinkedRecords('parents');
      if (parents != null) {
        expect(parents[0]?.getValue('name')).toEqual('Laius');
      } else {
        throw new Error('parents should not be null');
      }

      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      if (updatableData.node2 != null) {
        if (updatableData.node2.__typename === 'User') {
          updatableData.node2.parents = [];
          expect(updatableData.node2.parents).toEqual([]);
          expect(
            store
              .getRoot()
              ?.getLinkedRecord('node', {id: '5'})
              ?.getLinkedRecords('parents'),
          ).toEqual([]);
        }
      } else {
        throw new Error('Expected node2 to exist');
      }
    });

    const source = environment.getStore().getSource();
    const selector = operation.fragment;
    const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
      .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
    expect(readOnlyData.node2?.parents).toEqual([]);
  });

  it('throws if you assign null or undefined to a plural linked field', () => {
    environment.commitPayload(operation, {
      me: null,
      node: {
        __typename: 'User',
        id: '4',
        name: 'Gaius Julius Caesar',
      },
      node2: {
        __typename: 'User',
        id: '5',
        name: 'Gaius Julius Caesar Octavianus',
        parents: [
          {
            __typename: 'User',
            id: '467',
            name: 'Gaius Octavius',
            parents: [],
          },
        ],
      },
    });

    commitLocalUpdate(environment, store => {
      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      const source = environment.getStore().getSource();
      const selector = operation.fragment;
      const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
        .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);

      if (updatableData.node2 != null) {
        if (updatableData.node2.__typename === 'User') {
          if (readOnlyData.node != null) {
            if (readOnlyData.node.__typename === 'User') {
              expect(() => {
                /* eslint-disable-next-line ft-flow/no-flow-fix-me-comments */
                // $FlowFixMe
                updatableData.node2.parents = null;
              }).toThrowError();
            }
          }
        } else {
          throw new Error('Expected node2 to have typename User');
        }
      } else {
        throw new Error('Expected node2 to exist');
      }
    });
  });

  it('lets you access fields deeply', () => {
    environment.commitPayload(operation, {
      me: null,
      node: {
        __typename: 'User',
        id: '4',
        name: 'Yancy Fry',
      },
      node2: {
        __typename: 'User',
        id: '5',
        name: 'Philip Fry',
        parents: [
          {
            __typename: 'User',
            id: '4',
            name: 'Yancy Fry',
            parents: [
              {
                id: '5',
              },
            ],
          },
        ],
      },
    });
    commitLocalUpdate(environment, store => {
      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      if (updatableData.node2 != null) {
        if (updatableData.node2.__typename === 'User') {
          expect(updatableData.node2?.parents[0].parents[0].name).toBe(
            'Philip Fry',
          );
        } else {
          throw new Error('Expected node2 to be a User');
        }
      } else {
        throw new Error('Expected node2 to exist');
      }
    });
  });

  it('lets you use variables', () => {
    environment.commitPayload(operation, {
      me: null,
      node: {
        id: '4',
        __typename: 'Metahuman',
      },
      node2: {
        id: '5',
        __typename: 'Page',
      },
    });

    commitLocalUpdate(environment, store => {
      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery2,
        {id: '4'},
      ).updatableData;
      expect(updatableData.node?.__typename).toBe('Metahuman');

      const updatableData2 = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery2,
        {id: '5'},
      ).updatableData;
      expect(updatableData2.node?.__typename).toBe('Page');

      const updatableData3 = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery2,
        {id: '42'},
      ).updatableData;
      expect(updatableData3.node).toBe(undefined);
    });
  });

  it('should ignore extra variables, though flow will complain', () => {
    environment.commitPayload(operation, {
      me: null,
      node: {
        id: '4',
        __typename: 'Metahuman',
      },
      node2: null,
    });

    commitLocalUpdate(environment, store => {
      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery2,
        // $FlowFixMe[prop-missing] That's the point
        {id: '4', foo: 'bar'},
      ).updatableData;
      expect(updatableData.node?.__typename).toBe('Metahuman');
    });
  });

  it('does not throw when accessing a client extension field', () => {
    environment.commitPayload(operation, {
      me: {
        __typename: 'User',
        id: '4',
        name: 'Mark',
        author: {
          id: '5',
          client_best_friend: {
            id: '6',
            name: 'Sheryl',
          },
          client_nickname: 'Zucc',
        },
      },
      node: null,
      node2: null,
    });

    commitLocalUpdate(environment, store => {
      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;
      expect(() => {
        // The author field contains client_best_friend, which is a client extension
        updatableData.me?.author;
      }).not.toThrowError();
      expect(() => {
        // The author field contains client_nickname, which is a client extension
        updatableData.me?.author2;
      }).not.toThrowError();
    });
  });

  it('lets you update client extension scalar fields', () => {
    environment.commitPayload(operation, {
      me: {
        __typename: 'User',
        id: '4',
        name: 'Mark',
        author: {
          id: '5',
          client_best_friend: {
            id: '6',
            name: 'Sheryl',
          },
          client_nickname: 'Zucc',
        },
      },
      node: null,
      node2: null,
    });

    commitLocalUpdate(environment, store => {
      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      if (updatableData.me?.author2 != null) {
        updatableData.me.author2.client_nickname = 'Mr. Right';
      } else {
        throw new Error('Expected author to exist');
      }
      expect(updatableData.me?.author2?.client_nickname).toBe('Mr. Right');
    });
    const source = environment.getStore().getSource();
    const selector = operation.fragment;
    const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
      .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
    expect(readOnlyData.me?.author?.client_nickname).toBe('Mr. Right');
  });

  it('lets you navigate through client extension fields and update nested scalar fields', () => {
    environment.commitPayload(operation, {
      me: {
        __typename: 'User',
        id: '4',
        name: 'Mark',
        author: {
          id: '5',
          client_best_friend: {
            id: '6',
            name: 'Sheryl',
          },
          client_nickname: 'Zucc',
        },
      },
      node: null,
      node2: null,
    });

    commitLocalUpdate(environment, store => {
      const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
        updatableQuery,
        {},
      ).updatableData;

      if (updatableData.me?.author?.client_best_friend != null) {
        updatableData.me.author.client_best_friend.name = 'Mr. Right';
      } else {
        throw new Error('Expected author to exist');
      }
      expect(updatableData.me?.author?.client_best_friend.name).toBe(
        'Mr. Right',
      );
    });

    const source = environment.getStore().getSource();
    const selector = operation.fragment;
    const readOnlyData = ((RelayReader.read(source, selector) // $FlowFixMe[unclear-type] Just to cast it to a better type!
      .data: any): readUpdatableQueryEXPERIMENTALTestRegularQuery['response']);
    expect(readOnlyData.me?.author?.client_best_friend?.name).toBe('Mr. Right');
  });

  describe('with custom scalars', () => {
    it('should update custom scalar field', () => {
      // Read initial data
      const readableQuery = graphql`
        query readUpdatableQueryEXPERIMENTALTest2Query {
          updatable_scalar_field
        }
      `;
      const updateableQuery = graphql`
        query readUpdatableQueryEXPERIMENTALTest1Query @updatable {
          updatable_scalar_field
        }
      `;

      const operationDescriptor = createOperationDescriptor(readableQuery, {});
      const value = environment.lookup(operationDescriptor.fragment);
      expect(value.data).toEqual({
        updatable_scalar_field: undefined,
      });

      function updateCustomScalar(newValue: OpaqueScalarType) {
        commitLocalUpdate(environment, store => {
          const updatableData = store.readUpdatableQuery_EXPERIMENTAL(
            updateableQuery,
            {},
          ).updatableData;
          updatableData.updatable_scalar_field = newValue;
        });
      }

      // Update custom scalar field
      updateCustomScalar(
        createOpaqueScalarTypeValue('Alice', () => {
          throw new Error('Hello, Alice');
        }),
      );

      // Validate the updated value
      const updatedData = environment.lookup(operationDescriptor.fragment).data;
      expect(updatedData).toEqual({
        updatable_scalar_field: {
          name: 'Alice',
          callback: expect.any(Function),
        },
      });
      expect(() => {
        // $FlowFixMe[incompatible-use] we just checked that callback is a function
        updatedData.updatable_scalar_field.callback();
      }).toThrow('Hello, Alice');

      // Update custom value one more time
      updateCustomScalar(createOpaqueScalarTypeValue('Bob', jest.fn()));
      expect(environment.lookup(operationDescriptor.fragment).data).toEqual({
        updatable_scalar_field: {
          name: 'Bob',
          callback: expect.any(Function),
        },
      });
    });
  });
});
